from web3 import Web3

from actions_lib.actions.transfer.transfer import ALLOWANCE_ABI, CHAIN_ID_MAPPING, ERC20_APPROVE_ABI, convert_to_wei, get_web3, TOKEN_MAPPING
from actions_lib.actions.type import Action, ActionData

SUPPORTED_TOKENS = {
    "eth": {
        "usdc": "0xd10519Aa03FE7Ffb95189B52B74F94Fb33E2C88a",  # testnet
        # "usdc": "0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48",  # mainnet
    },
    "base": {
        "usdc": "0x45b58118a5374dccf7fd6f3f66c66278ab7473d9", # testnet
        # "usdc": "0x833589fCD6eDb6E08f4c7C32D4f71b54bdA02913", # mainnet
    }
}

CHAIN_NAME_MAPPING = {
    "eth": "ETH",
    "base": "Base"
}

def authorize_token(token, amount, step: any, **kwargs):
    redis_client = kwargs.get("redis_client")
    executor = kwargs.get("executor")
    spender_info = kwargs.get('spender')

    active_chain_key = f"user:active_chain:{executor}"
    active_chain = redis_client.get(active_chain_key) or "base"
    chain_raw = active_chain.lower()
    providers = kwargs.get("providers")

    spender_address = spender_info[chain_raw]['control_address']

    w3 = get_web3(providers, chain_raw)
    amount_to_transfer = convert_to_wei(float(amount), chain_raw, token.lower())
    
    token_address = TOKEN_MAPPING[chain_raw][token.lower()]['address']
    if token_address is not None:
        token_address = w3.to_checksum_address(token_address)
    action_params = {
        "func": "approve",
        "chainId": CHAIN_ID_MAPPING[chain_raw],
        'contract': token_address,
        '_spender': spender_address,
        '_value': amount_to_transfer,
        'abi': ERC20_APPROVE_ABI
    }
    action_data = ActionData(func='', group='', params=action_params)
    action = Action(msg=None, type='wallet', data=action_data.__dict__)
    
    return {'result': {"code": 200, "content": "Authorization parameters are ready"}, 'action': action.__dict__, 'next_action': None }

def show_authorized_tokens(step: any, **kwargs):
    spender_info = kwargs.get('spender')
    providers = kwargs.get('providers')
    executor = kwargs.get("executor")

    result = fetch_approved_amounts(executor, providers, spender_info)
    return { 'result': { 'code': 200, 'content': result }, 'action': None, 'next_action': None }

def get_allowance(providers, chain, token_address, owner, spender):
    web3 = get_web3(providers, chain)
    token_address = Web3.to_checksum_address(token_address)
    contract = web3.eth.contract(address=token_address, abi=ALLOWANCE_ABI)
    spender = Web3.to_checksum_address(spender)
    owner = Web3.to_checksum_address(owner)
    allowance = contract.functions.allowance(owner, spender).call()
    return allowance

def fetch_approved_amounts(owner, providers, spender_info):
    """Iterate over SUPPORTED_TOKENS to fetch the approved allowance for each token and return a Markdown table."""
    markdown_table = "| Token | Allowance |\n|-------|-----------|\n"
    for chain, tokens in SUPPORTED_TOKENS.items():
        for token_name, token_address in tokens.items():
            try:
                spender = spender_info.get(chain.lower()).get('control_address')
                allowance = get_allowance(providers, chain, token_address, owner, spender)
                allowance_in_readable_format = Web3.from_wei(allowance, 'mwei')  # Convert to a readable unit
                # Add a row to the Markdown table
                markdown_table += f"| {CHAIN_NAME_MAPPING[chain]} ({token_name.upper()}) | {allowance_in_readable_format} |\n"
            except Exception as e:
                print(f"Error fetching allowance for {token_name} on {chain}: {e}")
                markdown_table += f"| {CHAIN_NAME_MAPPING[chain]} ({token_name.upper()}) | Error |\n"

    return markdown_table



from typing import Tuple, Dict
from web3 import Web3
from ens import ENS
import logging

logger = logging.getLogger(__name__)

def default_if_none(value, default="NULL"):
    return default if value is None else value

def redis_add_contact(redis_client, user: str, name: str, address: str, ens_name: str = None) -> Tuple[int, str]:
    """
    Add a contact to the Redis database. 

    :param redis_client: The Redis client instance.
    :param user: The user identifier.
    :param name: The contact name.
    :param address: The contact Ethereum address.
    :param ens_name: The ENS name if available.
    :return: Tuple with status code and message.
    """
    if not Web3.is_address(address):
        return 410, f"Input address: {address} is not a valid Ethereum address"
    
    lower_key = f"{user}_{name}".lower()
    
    # Check if the contact already exists
    if redis_client.hget(f"contact:{lower_key}", "address"):
        return 420, f"Already added {name} to contacts"
    
    redis_client.hset(f"contact:{lower_key}", mapping={
        "address": address,
        "ens_name": default_if_none(ens_name)
    })
    
    return 200, f"Added {name} ({address}) to contacts successfully"

def redis_edit_contact(redis_client, user: str, name: str, address: str, ens_name: str = None) -> Tuple[int, str]:
    """
    Edit an existing contact in the Redis database.

    :param redis_client: The Redis client instance.
    :param user: The user identifier.
    :param name: The contact name.
    :param address: The new contact Ethereum address.
    :param ens_name: The new ENS name if available.
    :return: Tuple with status code and message.
    """
    if not Web3.is_address(address):
        return 410, f"Input address: {address} is not a valid Ethereum address"
    
    lower_key = f"{user}_{name}".lower()
    contact_key = f"contact:{lower_key}"

    # Check if the contact exists
    if not redis_client.exists(contact_key):
        return 404, f"Contact {name} not found"
    
    # Update the contact information
    redis_client.hset(contact_key, mapping={
        "address": address,
        "ens_name": default_if_none(ens_name)
    })
    
    return 200, f"Updated {name} ({address}) in contacts successfully"

def redis_delete_contact(redis_client, user: str, name: str) -> Tuple[int, str]:
    """
    Delete an existing contact from the Redis database.

    :param redis_client: The Redis client instance.
    :param user: The user identifier.
    :param name: The contact name to be deleted.
    :return: Tuple with status code and message.
    """
    lower_key = f"{user}_{name}".lower()
    contact_key = f"contact:{lower_key}"

    # Check if the contact exists
    if not redis_client.exists(contact_key):
        return 404, f"Contact {name} not found"
    
    # Delete the contact
    deleted = redis_client.delete(contact_key)
    
    if deleted:
        return 200, f"Deleted contact {name} successfully"
    else:
        return 500, f"Failed to delete contact {name}"

def get_contact_address(redis_client, user: str, contact: str, w3: Web3) -> Tuple[int, str]:
    """
    Retrieve a contact address. If the contact is an ENS name, resolve it using Web3.

    :param redis_client: The Redis client instance.
    :param user: The user identifier.
    :param contact: The contact name or Ethereum address.
    :param w3: The Web3 instance.
    :return: Tuple with status code and address or error message.
    """
    if Web3.is_address(contact):
        return 200, w3.to_checksum_address(contact)
    
    lower_key = f"{user}_{contact}".lower()
    address = redis_client.hget(f"contact:{lower_key}", "address")

    if not address and contact.endswith(".eth"):
        try:
            ns = ENS.from_web3(w3)
            address = ns.address(contact)
            logger.info(f"Resolved ENS: {contact} -> {address}")
            if not address:
                return 404, f"Cannot find {contact} on ENS, or the address format is incorrect."
        except Exception as e:
            logger.error(f"ENS resolution error: {e}")
            return 500, str(e)
    elif not address:
        return 404, f"Cannot find {contact} in contacts, or the address format is incorrect."
    
    return 200, w3.to_checksum_address(address)

def redis_show_contact(redis_client, user: str, name: str = None) -> Tuple[int, str]:
    """
    Display all contacts for a user in markdown format.

    :param redis_client: The Redis client instance.
    :param user: The user identifier.
    :param name: Optional contact name to filter.
    :return: Tuple with status code and markdown result or error message.
    """
    try:
        contact_mapping = get_user_contacts_mapping(redis_client, user)
        if name:
            contact_mapping = {name: contact_mapping.get(name, None)}
            if not contact_mapping[name]:
                return 404, f"Cannot find {name} in contacts."
        result = generate_markdown(contact_mapping)
    except Exception as e:
        logger.error(f"Error fetching contacts: {e}")
        return 500, str(e)
    
    return 200, result

def get_user_contacts_mapping(redis_client, user: str) -> Dict[str, Dict[str, str]]:
    """
    Retrieve the contact mapping for a specific user.

    :param redis_client: The Redis client instance.
    :param user: The user identifier.
    :return: Dictionary of contact names with addresses and ENS names.
    """
    contact_mapping = {}
    for key in redis_client.scan_iter(match=f"contact:{user.lower()}_*"):
        try:
            name = key.split('_')[-1]
            address = redis_client.hget(key, 'address')
            ens_name = redis_client.hget(key, 'ens_name')
            
            if address:
                contact_mapping[name] = {
                    "address": address,
                    "ens_name": default_if_none(ens_name)
                }
        except Exception as e:
            print(f"Error fetching contact: {e}", flush=True)
    
    return contact_mapping

def generate_markdown(mapping: Dict[str, Dict[str, str]]) -> str:
    """
    Generate a markdown table from a mapping of contacts.

    :param mapping: Dictionary of contact names to details.
    :return: Markdown formatted string.
    """
    markdown_lines = ["| Name | Address | ENS |", "|------|---------|----------|"]
    for name, details in mapping.items():
        address = default_if_none(details.get('address'))
        ens_name = default_if_none(details.get('ens_name'))
        markdown_lines.append(f"| {name} | {address} | {ens_name} |")
    
    return "\n".join(markdown_lines)


import time
from typing import Tuple, Dict, Any
from web3 import HTTPProvider, Web3
from eth_utils.abi import function_abi_to_4byte_selector, collapse_if_tuple
from actions_lib.actions.consts import CHAIN_ID_MAPPING, TOKEN_MAPPING
from actions_lib.actions.type import Action, ActionData
from actions_lib.utils.contact_tool import get_contact_address
from cdp import *

# ABIs
ALLOWANCE_ABI = """[{"constant":true,"inputs":[{"name":"_owner","type":"address"},{"name":"_spender","type":"address"}],"name":"allowance","outputs":[{"name":"","type":"uint256"}],"payable":false,"stateMutability":"view","type":"function"}]"""

CONTROL_CONTRACT_ABI = [
    {
        "inputs": [
            {"internalType": "address", "name": "from", "type": "address"},
            {"internalType": "address", "name": "to", "type": "address"},
            {"internalType": "uint256", "name": "value", "type": "uint256"},
        ],
        "name": "controlTransferFrom",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {"inputs": [{"internalType": "address", "name": "owner_", "type": "address"}], "name": "setOwner", "outputs": [], "stateMutability": "nonpayable", "type": "function"},
    {"inputs": [{"internalType": "address", "name": "token_", "type": "address"}], "name": "setToken", "outputs": [], "stateMutability": "nonpayable", "type": "function"},
    {"inputs": [], "stateMutability": "nonpayable", "type": "constructor"},
    {"inputs": [{"internalType": "address", "name": "token_", "type": "address"}, {"internalType": "uint256", "name": "value", "type": "uint256"}], "name": "withdraw", "outputs": [], "stateMutability": "nonpayable", "type": "function"},
    {"stateMutability": "payable", "type": "receive"},
    {"inputs": [], "name": "TRANSFER", "outputs": [{"internalType": "bytes4", "name": "", "type": "bytes4"}], "stateMutability": "view", "type": "function"},
    {"inputs": [], "name": "TRANSFERFROM", "outputs": [{"internalType": "bytes4", "name": "", "type": "bytes4"}], "stateMutability": "view", "type": "function"},
]

ERC20_APPROVE_ABI = """[{"constant":false,"inputs":[{"name":"_spender","type":"address"},{"name":"_value","type":"uint256"}],"name":"approve","outputs":[{"name":"","type":"bool"}],"payable":false,"stateMutability":"nonpayable","type":"function"}]"""

ERC20_TRANSFER_ABI = """[{"constant":false,"inputs":[{"name":"recipient","type":"address"},{"name":"amount","type":"uint256"}],"name":"transfer","outputs":[{"name":"","type":"bool"}],"payable":false,"stateMutability":"nonpayable","type":"function"}]"""

def record_transaction(redis_client, user_id: str, chain: str, tx_hash: str) -> None:
    key = f"user:tx:{user_id}:{chain}"
    redis_client.rpush(key, tx_hash)
    print(f"Recorded transaction {tx_hash} on {chain} for user {user_id}")

def mpc_transfer(mpc_address: WalletAddress, destination, chain_raw, token_raw, token_address, amount):
    tx_hash = None
    try:
        if chain_raw == 'base' and token_raw == 'usdc':
            transfer = mpc_address.transfer(amount, 'usdc', destination, gasless=True)
            try:
                transfer.wait()
            except Exception as e:
                print(f"Transfer error: {e}", flush=True)

            timeout = 10  
            start_time = time.time()
            while str(transfer.status) not in ['complete','failed']:
                if time.time() - start_time > timeout:
                    print("Timeout: Transaction status check exceeded 10 seconds.", flush=True)
                    break
                print(f"Waiting for transaction status: {transfer.status}", flush=True)
                transfer.reload()
                tx_hash = str(transfer.transaction_hash)
                time.sleep(1)

            if str(transfer.status) == 'complete':
                status = 200
                transfer_status = 1
                print(f"Transfer complete: {transfer.transaction_hash}", flush=True)
                print(f"Transfer successfully landed on-chain: {transfer.transaction_link}", flush=True)
            else:
                print(f"Error: {transfer.status}")
                status = 500
                transfer_status = 0
                print(f"Transfer failed on-chain: {transfer.transaction_link}", flush=True)

            info = {
                "content": generate_tx_link_content(str(transfer.status), transfer.transaction_link),
                "tx_hash": transfer.transaction_hash,
                "chain": chain_raw,
                "status": transfer_status,
                "code": status
            }

            return info
        else:
            # amount_to_transfer = convert_to_wei(float(amount), chain_raw, token_raw)
            # transfer = default_address.invoke_contract(
            #     contract_address=token_address,
            #     abi=ERC20_TRANSFER_ABI,
            #     method="transfer",
            #     args={"to": destination, "value": amount_to_transfer}
            # )
            info = {
                "content": f"In automatic mode, only USDC transfers on the Base mainnet are supported.",
                "code": 500
            }
            return info
    except Exception as e:
        print(f"Exception occurred: {e}, Tx hash: {tx_hash}", flush=True)
        info = {
            "content": f"{e}",
            "code": 500
        }
        return info

def get_web3(providers, chain):
    if providers[chain]['w3'] is None:
        providers[chain]['w3'] = Web3(HTTPProvider(providers[chain]['rpc']))
    return providers[chain]['w3']

def get_user_action_mode(redis_client, user_id):
    user_id = user_id.lower()
    action_mode_key = f"user:action_mode:{user_id}"
    mode = redis_client.get(action_mode_key)
    if mode:
        return mode
    else:
        return 'manual'

def convert_to_wei(amount, chain, token):
    decimal = TOKEN_MAPPING[chain][token.lower()]['decimal']
    amount = float(amount) * (10 ** decimal)
    return int(amount)

def convert_to_value(amount, chain, token):
    decimal = TOKEN_MAPPING[chain][token.lower()]['decimal']
    amount = amount / (10 ** decimal)
    return float(amount)

def get_token_address(chain, token):
    try:
        # Safely access nested keys using .get() to avoid KeyError
        address = TOKEN_MAPPING.get(chain, {}).get(token, {}).get("address")
        return address
    except Exception as e:
        return f"Get token address failed. Error: {str(e)}"

def transfer(receiver: str, amount: float, step: any, chain: str = "base", token: str = "usdc", **kwargs: Any) -> Dict[str, Any]:
    print(f"receiver: {receiver}, amount: {amount}, chain: {chain}, token: {token}, step: {step}")
    
    chain_raw = chain.lower()
    executor = kwargs.get("executor")
    redis_client = kwargs.get("redis_client")
    spender_info = kwargs.get('spender')
    providers = kwargs.get("providers")
    executor_balance = kwargs.get("executor_balance")
    run_mode = get_user_action_mode(redis_client, executor)

    if not providers or chain_raw not in providers:
        raise ValueError(f"Provider for chain {chain} not found")
    
    spender_address = spender_info[chain_raw]['control_address']
    control_contract = spender_info[chain_raw]['control_contract']
    sender_address = spender_info[chain_raw]['sender_address']
    private_key = spender_info[chain_raw]['sender_private_key']
    blockchain_explore = providers[chain_raw]['scan']
    w3 = get_web3(providers, chain_raw)
    eth_w3 = get_web3(providers, 'eth')
    token_raw = token.lower()
    token_address = get_token_address(chain_raw, token_raw)

    mpc_address = kwargs.get('mpc_address')
    if token_address is not None:
        token_address = w3.to_checksum_address(token_address)
    print(f"token address: {token_address}", flush=True)
    
    # print(f"w3: {w3}", flush=True)

    if run_mode.lower() == 'auto':
        return handle_auto_transfer(w3, eth_w3, mpc_address, executor, receiver, amount, token_address, spender_address, sender_address, control_contract, private_key, redis_client, chain_raw, token_raw, blockchain_explore, executor_balance)
    else:
        return handle_manual_transfer(w3, eth_w3, executor, receiver, amount, token_address, redis_client, chain_raw, token_raw)

def handle_auto_transfer(w3: Web3, eth_w3, mpc_address, executor, receiver, amount, token_address, spender_address, sender_address, control_contract, private_key, redis_client, chain_raw, token_raw, blockchain_explore, executor_balance):
    
    if token_raw == 'usdc' and chain_raw == 'base':
        amount_to_transfer = convert_to_wei(float(amount), chain_raw, token_raw)
        # try:
        #     decrease_token_amount(redis_client, executor, chain_raw, token_raw, amount_to_transfer)
        # except Exception as e:
        #     print(f"Exception occurred: {e}", flush=True)
        #     return {'result': {"content": f"{e}", "code": 500}, 'action': None, 'next_action': None}
        try:
            code, receiver_address = get_contact_address(redis_client, executor, receiver, eth_w3)
            if code != 200:
                return {'result': {"content": f"Cannot find {receiver} on contacts", "code": 420}, 'action': None, 'next_action': None}

            to = w3.to_checksum_address(receiver_address)
            executor = w3.to_checksum_address(executor)

            print(f"to: {to}, executor: {executor}", flush=True)
            info = mpc_transfer(mpc_address, to, chain_raw, token_raw, token_address, amount)
            # if info['code'] != 200:
            #     increase_token_amount(redis_client, executor, chain_raw, token_raw, amount_to_transfer)
            return {'result': info, 'action': None, 'next_action': None}
        except Exception as e:
            # increase_token_amount(redis_client, executor, chain_raw, token_raw, amount_to_transfer)
            return handle_transaction_error(e, control_contract, w3)
    else:
        token = token_raw.upper()
        if token_raw in ['btc', 'cbbtc', 'wbtc']:
            if chain_raw == 'base':
                token = 'cbBTC'
            else:
                token = 'WBTC'
        return {'result': {"code": 500, "content": f"Transfers of the current currency({token}) are not supported in automatic mode."}, 'action': None, 'next_action': None}
        

def handle_manual_transfer(w3: Web3, eth_w3, executor, receiver, amount, token_address, redis_client, chain_raw, token_raw):
    active_chain_key = f"user:active_chain:{executor}"
    active_chain = redis_client.get(active_chain_key) or "base"
    print(f"active_chain_key: {active_chain_key}, active_chain: {active_chain}", flush=True)
    
    active_chain_id = CHAIN_ID_MAPPING[active_chain.lower()]
    param_chain_id = CHAIN_ID_MAPPING[chain_raw]
    
    if param_chain_id != active_chain_id:
        return handle_incorrect_network(w3, eth_w3, executor, receiver, amount, chain_raw, token_raw, redis_client)
    
    code, receiver_address = get_contact_address(redis_client, executor, receiver, eth_w3)
    if code != 200:
        return {'result': {"content": f"Cannot find {receiver} on contact", "code": 420}, 'action': None, 'next_action': None}

    amount_to_transfer = convert_to_wei(float(amount), chain_raw, token_raw)
    to = w3.to_checksum_address(receiver_address)
    params = {
        'from': executor,
        'to': to,
        'amount': amount_to_transfer,
        'token': token_address,
        'chain': param_chain_id
    }
    
    if token_address is None:
        # token_address = TOKEN_MAPPING[chain_raw]['usdc']['address']
        params = {
            'to': to,
            'amount': amount_to_transfer,
            'chainId': CHAIN_ID_MAPPING[chain_raw],
            'contract': None,
            'func': 'transferETH',
            'abi': None
        }
    else:
        params = {
            'recipient': to,
            'amount': amount_to_transfer,
            'chainId': CHAIN_ID_MAPPING[chain_raw],
            'contract': token_address,
            'func': 'transfer',
            'abi': ERC20_TRANSFER_ABI
        }
    action_data = ActionData(func='transfer', group='transfer', params=params)
    action = Action(msg=None, type='wallet', data=action_data.__dict__)
    return {'result': {"code": 200, "content": "Manual transfer parameters are constructed"}, 'action': action.__dict__, 'next_action': None}

def handle_insufficient_allowance(w3, eth_w3, executor, receiver, amount, spender_address, token_address, chain_raw, token_raw, redis_client):
    amount_to_transfer = convert_to_wei(float(amount), chain_raw, token_raw)
    action_params = {
        "func": "approve",
        "chainId": CHAIN_ID_MAPPING[chain_raw],
        'contract': token_address,
        '_spender': spender_address,
        '_value': amount_to_transfer,
        'abi': ERC20_APPROVE_ABI
    }
    action_data = ActionData(func='', group='', params=action_params)
    action = Action(msg=None, type='wallet', data=action_data.__dict__)

    code, receiver_address = get_contact_address(redis_client, executor, receiver, eth_w3)
    if code != 200:
        return {'result': {"content": f"{receiver_address}", "code": 420}, 'action': None, 'next_action': None}

    next_params = {
        'receiver': w3.to_checksum_address(receiver_address),
        'amount': amount,
        'chain': chain_raw,
        'token': token_raw,
        'run_mode': 'auto'
    }
    next_action_data = ActionData(func='transfer', group='transfer', params=next_params)
    next_action = Action(msg=None, type='backend', data=next_action_data.__dict__)
    return {'result': {"code": 200, "content": "Insufficient allowance, proceeding with approval."}, 'action': action.__dict__, 'next_action': next_action.__dict__}

def handle_incorrect_network(w3, eth_w3, executor, receiver, amount, chain_raw, token_raw, redis_client):
    params = {"chainId": CHAIN_ID_MAPPING[chain_raw]}
    action_data = ActionData(func='switch_chain', group='', params=params)
    action = Action(msg=None, type='setting', data=action_data.__dict__)

    code, get_contact_res = get_contact_address(redis_client, executor, receiver, eth_w3)
    if code != 200:
        return {'result': {"content": get_contact_res, "code": code}, 'action': None, 'next_action': None}

    receiver_address = get_contact_res
    next_params = {
        'receiver': w3.to_checksum_address(receiver_address),
        'amount': amount,
        'chain': chain_raw,
        'token': token_raw,
        'run_mode': 'manual'
    }
    next_action_data = ActionData(func='transfer', group='transfer', params=next_params)
    next_action = Action(msg=None, type='backend', data=next_action_data.__dict__)

    return {'result': {'code': 200, 'content': 'Incorrect network, switching to the correct network for you.'}, 'action': action.__dict__, 'next_action': next_action.__dict__}

def maybe_send_transaction(w3: Web3, control_contract, sender_address, token, executor, to, amount_to_transfer, private_key, chain=None, executor_balance=None):
    nonce = w3.eth.get_transaction_count(sender_address)
    current_gas_price = w3.eth.gas_price
    increase = current_gas_price * 25  // 100
    new_gas_price = current_gas_price + increase
    tx = control_contract.functions.transferFromERC20(
        token, executor, to, amount_to_transfer).build_transaction({
            'from': sender_address,
            'nonce': nonce,
            'gasPrice': new_gas_price,
        })
    gas_estimate = w3.eth.estimate_gas(tx)
    tx['gas'] = gas_estimate
    gas_cost = gas_estimate * new_gas_price
    
    print(f"gas_estimate: {gas_estimate}, gas_cost: {gas_cost}, executor_balance: {executor_balance}")
    if executor_balance[chain] < gas_cost:
        return False, 0x0
    
    signed_tx = w3.eth.account.sign_transaction(tx, private_key)
    return True, w3.eth.send_raw_transaction(signed_tx.rawTransaction)

def handle_transaction_error(e, control_contract, w3):
    name, params = decode_custom_error(control_contract.abi, w3, str(e))
    if not name:
        message = f"Transfer failed with unknown error: {e}"
    else:
        message = f"Transfer error, error name: {name}, parameters: {params}"
    return {'result': {"content": message, "status": 0, "code": 500}, 'action': None, 'next_action': None}

def generate_markdown_content(status, hash_value, blockchain_explore):
    status_text = "successful" if status == 1 else "failed"
    return f"Transfer {status_text}. You can check the transaction on [blockchain explorer]({blockchain_explore}{hash_value})"


def generate_tx_link_content(status, tx_link):
    status_text = "successful" if status == 'complete' else "failed"
    return f"Transfer {status_text}. You can check the transaction on [blockchain explorer]({tx_link})"


def decode_custom_error(contract_abi, w3, error) -> Tuple[str, str]:
    # Parse error content, the error content must look like:
    # "Call method: submitServiceProof,xxxxx,error:('xxxxxx','xxxxxxx')"
    tmp_array = error.split(":")
    if len(tmp_array) != 3:
        return None, None
    param_str = tmp_array[2]
    param_str = param_str.replace("(","")
    param_str = param_str.replace(")","")
    param_str = param_str.replace(",","")
    param_str = param_str.replace("'","")
    errors = param_str.split()

    for error in [abi for abi in contract_abi if abi["type"] == "error"]:
        # Get error signature components
        name = error["name"]
        data_types = [collapse_if_tuple(abi_input) for abi_input in error.get("inputs", [])]
        error_signature_hex = function_abi_to_4byte_selector(error).hex()
        # Find match signature from error
        for error in errors:
            if error_signature_hex.casefold() == error[2:10].casefold():
                params = ','.join([str(x) for x in w3.codec.decode(data_types,bytes.fromhex(error[10:]))])
                #decoded = "%s(%s)" % (name , str(params))
                return name, params
    return None, None #try other contracts until result is not None since error may be raised from another called contract


from web3 import Web3
from ens import ENS
from actions_lib.utils.contact_tool import redis_add_contact, redis_edit_contact, redis_delete_contact, redis_show_contact
import logging

logger = logging.getLogger(__name__)

def resolve_ens_name(w3: Web3, address: str) -> str:
    """
    Resolves the ENS name for a given Ethereum address.

    :param w3: Web3 instance.
    :param address: Ethereum address.
    :return: ENS name if found, otherwise None.
    """
    try:
        ns = ENS.from_web3(w3)
        ens_name = ns.name(address)
        if ens_name:
            logger.info(f"Resolved ENS name for {address}: {ens_name}")
        else:
            logger.info(f"No ENS name found for the address {address}")
        return ens_name
    except Exception as e:
        logger.error(f"Error resolving ENS name for address {address}: {e}")
        return None

def add_contact(name: str, address: str, step: any, **kwargs) -> dict:
    """
    Adds a contact with the given name and address to the Redis database.

    :param name: Contact name.
    :param address: Contact Ethereum address.
    :param step: Step object in the workflow (details depending on object structure).
    :param kwargs: Additional parameters including redis_client, executor, and providers.
    :return: Result of the add operation.
    """
    redis_client = kwargs.get('redis_client')
    executor = kwargs.get('executor')
    providers = kwargs.get('providers')
    w3 = providers['ethereum']['w3']

    ens_name = resolve_ens_name(w3, address)
    code, res = redis_add_contact(redis_client, executor, name, address, ens_name)

    return {
        'result': {'code': code, 'content': res},
        'action': None,
        'next_action': None
    }

def edit_contact(name: str, address: str, step: any, **kwargs) -> dict:
    
    redis_client = kwargs.get('redis_client')
    executor = kwargs.get('executor')
    providers = kwargs.get('providers')
    w3 = providers['ethereum']['w3']
    ens_name = resolve_ens_name(w3, address)
    code, res = redis_edit_contact(redis_client, executor, name, address, ens_name)
    return {
        'result': {'code': code, 'content': res},
        'action': None,
        'next_action': None
    }

def delete_contact(name: str, step: any, **kwargs) -> dict:
    redis_client = kwargs.get('redis_client')
    executor = kwargs.get('executor')
    code, res = redis_delete_contact(redis_client, executor, name)
    return {
        'result': {'code': code, 'content': res},
        'action': None,
        'next_action': None
    }

def show_contact_by_name(name: str, step: any, **kwargs) -> dict:
    """
    Shows contact details by name.

    :param name: Contact name.
    :param step: Step object in the workflow (details depending on object structure).
    :param kwargs: Additional parameters including redis_client and executor.
    :return: Result of the show operation.
    """
    return show_contacts(name=name, **kwargs)

def show_all_contact(step: any, **kwargs) -> dict:
    """
    Shows all contacts for the executor.

    :param step: Step object in the workflow (details depending on object structure).
    :param kwargs: Additional parameters including redis_client and executor.
    :return: Result of the show operation.
    """
    return show_contacts(name=None, **kwargs)

def show_contacts(name: str = None, **kwargs) -> dict:
    """
    Shows contact details by name or shows all contacts if name is None.

    :param name: Contact name or None to show all.
    :param kwargs: Additional parameters including redis_client and executor.
    :return: Result of the show operation.
    """
    redis_client = kwargs.get('redis_client')
    executor = kwargs.get('executor')

    code, res = redis_show_contact(redis_client, executor, name)

    return {
        'result': {'code': code, 'content': res},
        'action': None,
        'next_action': None
    }


from .contact import add_contact, show_contact_by_name, show_all_contact, delete_contact, edit_contact

__all__ = [
    'add_contact',
    'show_contact_by_name',
    'show_all_contact',
    'delete_contact',
    'edit_contact'
]


from .transfer import transfer
from .contact import add_contact, show_all_contact, show_contact_by_name, delete_contact, edit_contact
from .type import Action, ActionData
from .user import show_account_info

__all__ = [
    'transfer',
    'add_contact',
    'show_all_contact',
    'show_contact_by_name',
    'Action',
    'ActionData',
    'delete_contact',
    'edit_contact',
    'show_account_info'
]


import os
import json

class JSONReader:
    def __init__(self, base_dir = None):
        if base_dir is None:
            # Use the default path inside the library
            base_dir = os.path.dirname(__file__)
        self.base_dir = base_dir

    def read_json(self, subdirectory_name, file_name, file_suffix = '.json'):
        """
        Read the content of a JSON file based on the subdirectory name.
        
        :param subdirectory_name: Name of the subdirectory
        :return: Content of the JSON file parsed as a Python object
        :raises FileNotFoundError: If the JSON file does not exist
        :raises ValueError: If the JSON content cannot be parsed
        """
        # Construct the subdirectory path
        subdirectory_path = self.base_dir
        if subdirectory_name is not None:
            subdirectory_path = os.path.join(self.base_dir, subdirectory_name)
        # Construct the JSON file path
        if file_suffix is not None:
            json_file_path = os.path.join(subdirectory_path, f"{file_name}{file_suffix}")
        else:
            json_file_path = os.path.join(subdirectory_path, f"{file_name}")
        
        # Check if the JSON file exists
        if not os.path.exists(json_file_path):
            raise FileNotFoundError(f"JSON file not found: {json_file_path}")
        
        # Read and return the JSON file content
        with open(json_file_path, 'r', encoding='utf-8') as json_file:
            try:
                data = json.load(json_file)
                return data
            except json.JSONDecodeError as e:
                raise ValueError(f"Error decoding JSON file {json_file_path}: {str(e)}")


from typing import Optional


class ActionData:

    func: str

    group: str

    params: dict

    def __init__(self, func, group, params) -> None:
        self.func = func
        self.group = group
        self.params = params

class Action:

    type: str

    msg: Optional[str] = None

    data: dict

    def __init__(self, type, msg, data) -> None:
        self.data = data
        self.type = type
        self.msg = msg




from web3 import Web3

from actions_lib.actions.consts import TOKEN_MAPPING

ERC20_ABI = [
    {
        "constant": True,
        "inputs": [{"name": "_owner", "type": "address"}],
        "name": "balanceOf",
        "outputs": [{"name": "balance", "type": "uint256"}],
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "decimals",
        "outputs": [{"name": "", "type": "uint8"}],
        "type": "function",
    },
]

def show_account_info(step: any, **kwargs):
    redis_client = kwargs.get('redis_client')
    executor = kwargs.get('executor')
    providers = kwargs.get("providers")
    w3 = providers['base']['w3']
    result = {}
    account_info = get_account_info(redis_client, executor)
    user_assets = get_user_assets(redis_client, executor, w3)
    result['remain_balance'] = account_info['remain_balance']
    result['address'] = executor
    result['assets'] = user_assets

    res = generate_complete_markdown(executor, account_info['remain_balance'], user_assets)
    return {
        'result': {'code': 200, 'content': res},
        'action': None,
        'next_action': None
    } 

def get_account_info(redis_client, user_id):
    user_id = user_id.lower()
    print(f"user_id: {user_id}")
    account_key = f"user:account:{user_id}"
    default_values = {
        'charge_total': '0',
        'action_fee': '0',
        'tx_fee': '0'
    }
    account_info = redis_client.hgetall(account_key)
    default_values.update(account_info)
    result = {k: int(default_values.get(k, '0')) for k in ['charge_total', 'action_fee', 'tx_fee']}
    extra_fund = get_extra_funds(redis_client, user_id)
    
    result['remain_balance'] = result['charge_total'] - result['action_fee'] - result['tx_fee'] + extra_fund
    return result

def get_extra_funds(redis_client, account_id):
    account_id = account_id.lower()
    extra_funds = redis_client.hget(f'user_funds:{account_id}', 'extra_funds')
    return int(extra_funds or 0)

def get_base_usdc_balance(address: str, w3: Web3) -> float:
    address = Web3.to_checksum_address(address)
    # usdc_address = Web3.to_checksum_address(TOKEN_MAPPING['base']['usdc']['address'])
    usdc_address = Web3.to_checksum_address("0x036CbD53842c5426634e7929541eC2318f3dCF7e")
    usdc_contract = w3.eth.contract(
        address=usdc_address,
        abi=ERC20_ABI
    )
    raw_balance = usdc_contract.functions.balanceOf(address).call()
    return raw_balance

def get_user_assets(redis_client, user_id, w3):
    result = []
    chain = 'base-sepolia'
    user_id = user_id.lower()
    chain = chain.lower()   
    key = f"mpc_address:{user_id}:{chain}"
    
    redis_data = redis_client.hgetall(key)
    if not redis_data:
        result.append({'chain': 'Base', 'token': 'USDC', 'amount': 0 })
        return result
    else:
        usdc_balance = get_base_usdc_balance(redis_data['wallet_address'], w3)
        result.append({'chain': 'Base', 'token': 'USDC', 'amount': usdc_balance})
        return result

def generate_complete_markdown(user_address, ai_balance, assets):
    """
    Generate a complete Markdown report with user info and assets.

    Args:
        user_address (str): The user's Ethereum address.
        ai_balance (float): The remaining AI balance.
        assets (list): A list of dictionaries with 'chain', 'token', and 'amount'.

    Returns:
        str: A complete Markdown report.
    """
    # Header with user address and AI balance
    header = (
        f"**Account address:** `{user_address}`\n"
        f"**Inference fee remain balance:** ${convert_to_value(ai_balance, 6)}\n"
        f"**Assets:**\n\n"
    )
    
    # Define the table structure
    table_header = "| Chain | Token | Amount |\n|-------|-------|--------|\n"
    rows = [f"| {asset['chain']} | {asset['token'].upper()} | {convert_to_value(asset['amount'], 6)} |" for asset in assets]
    table = table_header + "\n".join(rows)
    
    # Combine the header and table into the final Markdown output
    markdown_output = header + table
    return markdown_output

def convert_to_value(amount, decimal):
    amount = amount / (10 ** decimal)
    return float(amount)